"""Parse metadata from .dist-info directories in a wheelhouse."""
# flake8: noqa: E501

import click
import hashlib
import json
import os.path
import requests
import tempfile
import pkg_resources

from pypi2nix.utils import TO_IGNORE, safe, cmd


EXTENSIONS = ['.tar.gz', '.tar.bz2', '.tar', '.zip', '.tgz']
INDEX_URL = "https://pypi.io/pypi"
INDEX_URL = "https://pypi.python.org/pypi"


def find_homepage(item):
    homepage = ''
    if 'extensions' in item and \
            'python.details' in item['extensions'] and \
            'project_urls' in item['extensions']['python.details']:
        homepage = item['extensions']['python.details']['project_urls'].get('Home', '')
    return homepage


def extract_deps(metadata):
    """Get dependent packages from metadata.

    Note that this is currently very rough stuff. I consider only the
    first 'requires' dataset in 'run_requires'. Other requirement sets
    like 'test_requires' are completely ignored.
    """
    deps = []
    if 'run_requires' in metadata:
        for item in metadata['run_requires']:
            if 'requires' in item:
                for line in item['requires']:
                    components = line.split()

                    dep = components[0]
                    dep = dep.split("==")[0]
                    dep = dep.split(">=")[0]
                    dep = dep.split("<=")[0]
                    dep = dep.split("<")[0]
                    dep = dep.split(">")[0]

                    if dep.lower() in TO_IGNORE:
                        continue

                    if '[' in dep:
                        deps.append(dep.split('[')[0])
                    else:
                        deps.append(dep)

    return list(set(deps))

all_classifiers = {
    'License :: Aladdin Free Public License (AFPL)': None,
    'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication': None,
    'License :: DFSG approved': None,
    'License :: Eiffel Forum License (EFL)': None,
    'License :: Free For Educational Use': None,
    'License :: Free For Home Use': None,
    'License :: Free for non-commercial use': None,
    'License :: Freely Distributable': None,
    'License :: Free To Use But Restricted': None,
    'License :: Freeware': None,
    'License :: Netscape Public License (NPL)': None,
    'License :: Nokia Open Source License (NOKOS)': None,
    'License :: OSI Approved': None,
    'License :: OSI Approved :: Academic Free License (AFL)': 'licenses.afl21',
    'License :: OSI Approved :: Apache Software License': None,
    'License :: OSI Approved :: Apple Public Source License': None,
    'License :: OSI Approved :: Artistic License': 'licenses.artistic2',
    'License :: OSI Approved :: Attribution Assurance License': None,
    'License :: OSI Approved :: BSD License': 'licenses.bsdOriginal',
    'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)': None,
    'License :: OSI Approved :: Common Public License': 'licenses.cpl10',
    'License :: OSI Approved :: Eiffel Forum License': 'licenses.efl20',
    'License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)': None,
    'License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)': None,
    'License :: OSI Approved :: GNU Affero General Public License v3': 'licenses.agpl3',
    'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)': 'licenses.agpl3Plus',
    'License :: OSI Approved :: GNU Free Documentation License (FDL)': 'licenses.fdl13',
    'License :: OSI Approved :: GNU General Public License (GPL)': 'licenses.gpl1',
    'License :: OSI Approved :: GNU General Public License v2 (GPLv2)': 'licenses.gpl2',
    'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)': 'licenses.gpl2Plus',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)': 'licenses.gpl3',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)': 'licenses.gpl3Plus',
    'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)': 'licenses.lgpl2',
    'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)': 'licenses.lgpl2Plus',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)': 'licenses.lgpl3',
    'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)': 'licenses.lgpl3Plus',
    'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)': 'licenses.lgpl2',
    'License :: OSI Approved :: IBM Public License': 'licenses.ipl10',
    'License :: OSI Approved :: Intel Open Source License': None,
    'License :: OSI Approved :: ISC License (ISCL)': 'licenses.isc',
    'License :: OSI Approved :: Jabber Open Source License': None,
    'License :: OSI Approved :: MIT License': 'licenses.mit',
    'License :: OSI Approved :: MITRE Collaborative Virtual Workspace License (CVW)': None,
    'License :: OSI Approved :: Motosoto License': None,
    'License :: OSI Approved :: Mozilla Public License 1.0 (MPL)': 'licenses.mpl10',
    'License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)': 'licenses.mpl11',
    'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)': 'licenses.mpl20',
    'License :: OSI Approved :: Nethack General Public License': None,
    'License :: OSI Approved :: Nokia Open Source License': None,
    'License :: OSI Approved :: Open Group Test Suite License': None,
    'License :: OSI Approved :: Python License (CNRI Python License)': None,
    'License :: OSI Approved :: Python Software Foundation License': 'licenses.psfl',
    'License :: OSI Approved :: Qt Public License (QPL)': None,
    'License :: OSI Approved :: Ricoh Source Code Public License': None,
    'License :: OSI Approved :: Sleepycat License': 'licenses.sleepycat',
    'License :: OSI Approved :: Sun Industry Standards Source License (SISSL)': None,
    'License :: OSI Approved :: Sun Public License': None,
    'License :: OSI Approved :: University of Illinois/NCSA Open Source License': 'licenses.ncsa',
    'License :: OSI Approved :: Vovida Software License 1.0': 'licenses.vsl10',
    'License :: OSI Approved :: W3C License': 'licenses.w3c',
    'License :: OSI Approved :: X.Net License': None,
    'License :: OSI Approved :: zlib/libpng License': 'licenses.zlib',
    'License :: OSI Approved :: Zope Public License': 'licenses.zpl21',
    'License :: Other/Proprietary License': None,
    'License :: Public Domain': 'licenses.publicDomain',
    'License :: Repoze Public License': None,
}


def find_license(item):
    license = None

    classifiers = item.get('classifiers', [])

    # find first license classifier
    all_classifiers_keys = all_classifiers.keys()
    license_classifiers = [i for i in filter(
        lambda x: x in all_classifiers_keys,
        classifiers
    )]
    for license_classifier in license_classifiers:
        license_nix = all_classifiers[license_classifier]
        if license_nix is not None:
            license = license_nix
            break

    if license is None:
        license = item.get('license', '')

        if license in ['LGPL with exceptions or ZPL', 'ZPL 2.1']:
            license = "licenses.zpl21"
        elif license in ['MIT', 'MIT License',
                         'MIT or Apache License, Version 2.0',
                         'The MIT License'
        ]:
            license = "licenses.mit"
        elif license in ['BSD', 'BSD License', 'BSD-like',
                         'BSD or Apache License, Version 2.0'] or \
                license.startswith('BSD -'):
            license = "licenses.bsdOriginal"
        elif license in ['Apache 2.0', 'Apache License 2.0', 'Apache 2',
                         'Apache License, Version 2.0',
                         'Apache License Version 2.0']:
            license = "licenses.asl20"
        elif license in ['GNU Lesser General Public License (LGPL), Version 3',
                         'LGPL']:
            license = "licenses.lgpl3"
        elif license in ['MPL 2.0', 'MPL 2.0 (Mozilla Public License)', 'MPL-2.0']:
            license = "licenses.mpl20"
        elif license in ['Python Software Foundation License']:
            license = "licenses.psfl"

        else:
            if len(license_classifiers) > 0:
                license = license_classifiers[0]
            else:
                click.echo(
                    "WARNING: Couldn't recognize license `{}` for `{}`".format(
                        license, item.get('name')))

            license = '"{}"' .format(safe(license))

    return license


def process_metadata(wheel):
    """Find the actual metadata json file from several possible names.
    """
    for _file in ('metadata.json', 'pydist.json'):
        wheel_file = os.path.join(wheel, _file)
        if os.path.exists(wheel_file):
            with open(wheel_file) as f:
                metadata = json.load(f)
                if metadata['name'].lower() in TO_IGNORE:
                    return
                else:
                    return {
                        'name': metadata['name'],
                        'version': metadata['version'],
                        'deps': extract_deps(metadata),
                        'homepage': safe(find_homepage(metadata)),
                        'license': find_license(metadata),
                        'description': safe(metadata.get('summary', '')),
                    }
    raise click.ClickException(
        "Unable to find metadata.json/pydist.json in `%s` folder." % wheel)


def download_file(url, filename, chunk_size=2048):
    r = requests.get(url, stream=True, timeout=None)
    r.raise_for_status()  # TODO: handle this nicer

    with open(filename, 'wb') as fd:
        for chunk in r.iter_content(chunk_size):
            fd.write(chunk)


def find_release(wheel_cache_dir, wheel, wheel_data):

    wheel_release = None

    _release_version = wheel['version']
    _releases = wheel_data['releases'].get(wheel['version'], [])

    # sometimes version in release list is not exact match and we need to use
    # pkg_resources's parse_version function to detect which release list is
    # correct
    if not _releases:
        for _version, _releases_tmp in wheel_data['releases'].items():
            if pkg_resources.parse_version(wheel['version']) == \
               pkg_resources.parse_version(_version):
                _release_version = _version
                _releases = _releases_tmp
                break

    # sometimes for some unknown reason release data is under other version.
    # example: https://pypi.python.org/pypi/radiotherm/json
    if not _releases:
        _base_version = pkg_resources.parse_version(wheel['version']).base_version
        for _releases_tmp in wheel_data['releases'].values():
            for _release_tmp in _releases_tmp:
                for _ext in EXTENSIONS:
                    if _release_tmp['filename'].endswith(wheel['version'] + _ext):
                        _release_version = wheel['version']
                        _releases = [_release_tmp]
                        break
                    if _release_tmp['filename'].endswith(_base_version + _ext):
                        _release_version = _base_version
                        _releases = [_release_tmp]
                        break

    # a release can come in different formats. formats we care about are
    # listed in EXTENSIONS variable
    for _release in _releases:
        for _ext in EXTENSIONS:
            if _release['filename'].endswith(_ext):
                wheel_release = _release
                break
        if wheel_release:
            break

    if not wheel_release:
        raise click.ClickException(
            "Unable to find release for package {name} of version "
            "{version}".format(**wheel))

    release = dict()
    release['url'] = wheel_release['url']
    digests = wheel_release.get('digests')
    release['hash_type'] = 'sha256'
    if digests:
        release['hash_value'] = wheel_release['digests']['sha256']  # noqa
    else:
        # download file if it doens not already exists
        filename = os.path.join(
            wheel_cache_dir, wheel_release['filename'])
        if not os.path.exists(filename):
            download_file(wheel_release['url'], filename)

        # calculate sha256
        with open(filename, 'rb') as f:
            hash = hashlib.sha256(f.read())
        release['hash_value'] = hash.hexdigest()

    return release


def process_wheel(wheel_cache_dir, wheel, sources, verbose, index=INDEX_URL,
                  chunk_size=2048):
    """
    """

    if wheel['name'] in sources:
        release = dict()
        release['url'] = sources[wheel['name']]['url']
        release['hash_type'] = 'sha256'

        repo_type = sources[wheel['name']]['type']
        if repo_type == 'url':

            release['fetch_type'] = 'fetchurl'

            r = requests.get(release['url'], stream=True, timeout=None)
            r.raise_for_status()  # TODO: handle this nicer

            with tempfile.TemporaryFile() as fd:
                for chunk in r.iter_content(chunk_size):
                    fd.write(chunk)
                    fd.seek(0)
                    hash = hashlib.sha256(fd.read())

            release['hash_value'] = hash.hexdigest()

        elif repo_type == 'git':
            revision = ''
            if release['url'].startswith('git+'):
                release['url'] = release['url'][4:]
            if '@' in release['url']:
                release['url'], revision = release['url'].split('@')

            release['fetch_type'] = 'fetchgit'
            command = 'nix-prefetch-git {url} {revision}'.format(
                url=release['url'],
                revision=revision,
            )
            return_code, output = cmd(command, verbose != 0)
            if return_code != 0:
                raise click.ClickException("URL {url} for package {name} is not valid.".format(
                    url=release['url'],
                    name=wheel['name']
                ))
            for output_line in output.split('\n'):
                output_line = output_line.strip()
                if output_line.startswith('hash is '):
                    release['hash_value'] = output_line[len('hash is '):].strip()
                elif output_line.startswith('git revision is '):
                    release['rev'] = output_line[len('git revision is '):].strip()

            if release.get('hash_value', None) is None:
                raise click.ClickException('Could not determine the hash from ouput:\n{output}'.format(
                    output=output
                ))
            if release.get('rev', None) is None:
                raise click.ClickException('Could not determine the revision from ouput:\n{output}'.format(
                    output=output
                ))

        elif repo_type == 'hg':
            revision = ''
            if release['url'].startswith('hg+'):
                release['url'] = release['url'][3:]
            if '@' in release['url']:
                release['url'], revision = release['url'].split('@')

            release['fetch_type'] = 'fetchhg'
            command = 'nix-prefetch-hg {url} {revision}'.format(
                url=release['url'],
                revision=revision,
            )
            return_code, output = cmd(command, verbose != 0)
            if return_code != 0:
                raise click.ClickException("URL {url} for package {name} is not valid.".format(
                    url=release['url'],
                    name=wheel['name']
                ))
            HASH_PREFIX = 'hash is '
            REV_PREFIX = 'hg revision is '
            for output_line in output.split('\n'):
                print(output_line)
                output_line = output_line.strip()
                if output_line.startswith(HASH_PREFIX):
                    release['hash_value'] = output_line[len(HASH_PREFIX):].strip()
                elif output_line.startswith(REV_PREFIX):
                    release['rev'] = output_line[len(REV_PREFIX):].strip()

            if release.get('hash_value', None) is None:
                raise click.ClickException('Could not determine the hash from ouput:\n{output}'.format(
                    output=output
                ))
            if release.get('rev', None) is None:
                raise click.ClickException('Could not determine the revision from ouput:\n{output}'.format(
                    output=output
                ))

        elif repo_type == 'path':
            release['fetch_type'] = 'path'

        else:
            raise click.ClickException('Source type `{}` not implemented'.format(repo_type))

    else:
        url = "{}/{}/json".format(index, wheel['name'])
        r = requests.get(url, timeout=None)
        r.raise_for_status()  # TODO: handle this nicer
        wheel_data = r.json()

        if not wheel_data.get('releases'):
            raise click.ClickException(
                "Unable to find releases for packge {name}".format(**wheel))

        release = find_release(wheel_cache_dir, wheel, wheel_data)

    wheel.update(release)

    return wheel


def main(verbose, wheels, requirements_files, wheel_cache_dir, index=INDEX_URL,
         sources=dict()):
    """Extract packages metadata from wheels dist-info folders.
    """

    # get url's from requirements_files
    sources_urls = [i['url'] for i in sources.values()]
    for requirements_file in requirements_files:
        with open(requirements_file) as f:
            lines = f.readlines()
            for line in lines:
                line = line.strip()
                if line.startswith('-e '):
                    line = line[3:]

                if os.path.isdir(line) and line not in sources_urls:
                    raise click.ClickException(
                        "Source for path `%s` does not exists." % line
                    )

                elif (line.startswith('http://') or
                      line.startswith('https://') or
                      line.startswith('git+') or
                      line.startswith('hg+')):
                    try:
                        url, egg = line.split('#')
                        name = egg.split('egg=')[1]
                        if line.startswith('git+'):
                            sources[name] = dict(url=url, type='git')
                        elif line.startswith('hg+'):
                            sources[name] = dict(url=url, type='hg')
                        else:
                            sources[name] = dict(url=url, type='url')
                    except:
                        raise click.ClickException(
                            "Requirement starting with http:// or https:// "
                            "should end with #egg=<name>. Line `-e %s` does "
                            "not end with egg=<name>" % line
                        )

    output = ''
    metadata = []

    if verbose > 1:
        click.echo("-- sources ---------------------------------------------------------------")
        click.echo(json.dumps(sources, sort_keys=True, indent=4))
        click.echo("--------------------------------------------------------------------------")

    try:
        for wheel in wheels:

            output += '|-> from %s' % os.path.basename(wheel)
            if verbose != 0:
                click.echo('|-> from %s' % os.path.basename(wheel))

            wheel_metadata = process_metadata(wheel)
            if not wheel_metadata:
                continue

            if wheel_metadata['name'] in TO_IGNORE:
                if verbose != 0:
                    click.echo('    SKIPPING')
                continue

            if verbose > 1:
                click.echo("-- wheel_metadata --------------------------------------------------------")
                click.echo(json.dumps(wheel_metadata, sort_keys=True, indent=4))
                click.echo("--------------------------------------------------------------------------")

            metadata.append(
                process_wheel(wheel_cache_dir, wheel_metadata, sources,
                              verbose, index))
    except Exception as e:
        if verbose == 0:
            click.echo(output)
        raise e

    return metadata



import click
import glob
import os
import sys
import urllib

import pypi2nix.utils


HERE = os.path.dirname(__file__)


def main(verbose,
         requirements_files,
         project_dir,
         download_cache_dir,
         wheel_cache_dir,
         pip_build_dir,
         extra_build_inputs,
         python_version,
         nix_path=None,
         nix_shell='nix-shell',
         setup_requires=[],
         extra_env='',
         ):
    """Create a complete (pip freeze) requirements.txt and a wheelhouse from
       a user provided requirements.txt.
    """

    if nix_path:
        nix_path = ' '.join('-I {}'.format(i) for i in nix_path)
    else:
        nix_path = ''

    if extra_env:
        nix_instantiate = 'nix-instantiate'
        if os.path.exists(nix_shell):
            nix_instantiate = \
                os.path.abspath(os.path.join(
                    os.path.dirname(nix_path),
                    nix_instantiate
                ))
        command_env = '%s %s --eval --expr \'let pkgs = import <nixpkgs> {}; in "%s"\'' % (  # noqa
            nix_instantiate,
            nix_path,
            extra_env.replace('"', '\\"')
        )
        returncode, output = pypi2nix.utils.cmd(command_env, verbose != 0)
        if returncode != 0:
            if verbose == 0:
                click.echo(output)
            raise click.ClickException('Failed to interpret extra_env')
        extra_env = output.split('\n')[-2].strip()[1:-1]

    command = '{nix_shell} {nix_file} {nix_options} {nix_path} --show-trace --pure --run exit'.format(  # noqa
        nix_shell=nix_shell,
        nix_file=os.path.join(os.path.dirname(__file__), 'pip.nix'),
        nix_options=pypi2nix.utils.create_command_options(dict(
            requirements_files=requirements_files,
            project_dir=project_dir,
            download_cache_dir=download_cache_dir,
            wheel_cache_dir=wheel_cache_dir,
            pip_build_dir=pip_build_dir,
            extra_build_inputs=extra_build_inputs,
            extra_env=extra_env,
            python_version=python_version,
            setup_requires=setup_requires,
        )),
        nix_path=nix_path,
    )

    returncode, output = pypi2nix.utils.cmd(command, verbose != 0)
    if returncode != 0 or \
       output.endswith('ERROR: Failed to build one or more wheels'):
        if verbose == 0:
            click.echo(output)

        message = u'While trying to run the command something went wrong.'

        # trying to recognize the problem and provide more meanigful error
        # message
        no_matching_dist = "No matching distribution found for "
        if no_matching_dist in output:
            dist_name = output[
                output.find(no_matching_dist) + len(no_matching_dist):
            ]
            dist_name = dist_name[:dist_name.find(' (from')]
            message = (
                "Most likely `%s` package does not have source (zip/tar.bz) "
                "distribution." % dist_name
            )

        # if error is unknown then we ask to report an issue
        else:
            if click.confirm('Do you want to report above issue (a browser '
                             'will open with prefilled details of issue)?'):
                title = "Error when running pypi2nix command"
                body = "# Description\n\n<detailed description of error "
                "here>\n\n"
                body += "# Traceback \n\n```bash\n"
                body += "% pypi2nix --version\n"
                with open(os.path.join(HERE, 'VERSION')) as f:
                    body += f.read() + "\n"
                body += "% pypi2nix " + ' '.join(sys.argv[1:]) + "\n"
                body += output + "\n```\n"
                click.launch(
                    'https://github.com/garbas/pypi2nix/issues/new?%s' % (
                        urllib.parse.urlencode(dict(title=title, body=body))
                    )
                )

        raise click.ClickException(message)

    return (
        os.path.join(project_dir, 'requirements.txt'),
        glob.glob(os.path.join(project_dir, 'wheelhouse', '*.dist-info')),
    )



import os
import click

import pypi2nix.utils


def main(verbose,
         buildout_file,
         project_dir,
         buildout_cache_dir,
         extra_build_inputs,
         python_version,
         nix_path=None,
         nix_shell='nix-shell',
         setup_requires=[],
         ):
    """ Converts buildout.cfg specifiation into requirements.txt file
    """

    command = '{nix_shell} {nix_file} {options} {nix_path} -K --show-trace --pure --run exit'.format(  # noqa
        nix_shell=nix_shell,
        nix_file=os.path.join(os.path.dirname(__file__), 'buildout.nix'),
        options=pypi2nix.utils.create_command_options(dict(
            buildout_file=buildout_file,
            project_dir=project_dir,
            buildout_cache_dir=buildout_cache_dir,
            extra_build_inputs=extra_build_inputs,
            python_version=python_version,
            setup_requires=setup_requires,
        )),
        nix_path=nix_path \
            and ' '.join('-I {}'.format(i) for i in nix_path) \
            or ''
    )

    returncode, output = pypi2nix.utils.cmd(command, verbose != 0)
    if returncode != 0:
        if verbose == 0:
            click.echo(output)
        raise click.ClickException(
            u'While trying to run the command something went wrong.')

    return buildout_file and \
        os.path.join(project_dir,
                     'buildout_requirements.txt') \
        or None



import json
import shlex
import subprocess

import click

TO_IGNORE = [
    "pip",
    "setuptools",
    "wheel",
    "zc.buildout",
    "zc.recipe.egg",
]

PYTHON_VERSIONS = {
    "2.6": "python26Full",
    "2.7": "python27Full",
    "3.2": "python32",
    "3.3": "python33",
    "3.4": "python34",
    "3.5": "python35",
    "3.6": "python36",
    "3": "python3",
    "pypy": "pypy",
}


def pretty_option(option):
    if option is None:
        return ''
    else:
        return ' [value: {}]'.format(
            type(option) in [list, tuple] and
            ' '.join(option) or
            option)


def safe(string):
    return string.replace('"', '\\"')


def cmd(command, verbose=False, stderr=subprocess.STDOUT):

    if isinstance(command, str):
        command = shlex.split(command)

    if verbose:
        click.echo('|-> ' + ' '.join(command))

    p = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=stderr,
    )

    out = []
    while True:
        line = p.stdout.readline().decode()
        if line == '' and p.poll() is not None:
            break
        if line != '':
            if verbose:
                click.echo('    ' + line.rstrip('\n'))
            out.append(line)

    return p.returncode, '\n'.join(out)


def create_command_options(options):
    command_options = []
    for name, value in options.items():
        if isinstance(value, str):
            command_options.append('--argstr {} "{}"'.format(name, value))
        elif isinstance(value, list) or isinstance(value, tuple):
            value = "[ %s ]" % (' '.join(['"%s"' % x for x in value]))
            command_options.append("--arg {} '{}'".format(name, value))
    return ' '.join(command_options)


def args_as_list(inputs):
    return list(filter(
        lambda x: x != '',
        (' '.join(inputs)).split(' ')
    ))


def prefetch_git(url, rev=None):
    command = ['nix-prefetch-git', url]

    if rev is not None:
        command += ['--rev', rev]
    try:
        completed_proc = subprocess.run(
            command,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except FileNotFoundError:
        raise click.ClickException(
            'Could not find executable `nix-prefetch-git`.  '
            'Make sure it is installed correctly and available in '
            '$PATH.'
        )

    returncode = completed_proc.returncode

    if returncode != 0:
        raise click.ClickException(
            ('Could not fetch git repository at {url}, git returncode was '
             '{code}. stdout:\n{stdout}\nstderr:\n{stderr}').format(
                 url=url,
                 code=returncode,
                 stdout=completed_proc.stdout,
                 stderr=completed_proc.stderr,
            )
        )
    repo_data = json.loads(completed_proc.stdout)
    return repo_data



import subprocess
from urllib.parse import urldefrag, urlparse

import click

from .utils import cmd, prefetch_git


class UnsupportedUrlError(Exception):
    pass


class OverridesFile(object):
    def __init__(self, path):
        self.path = path

    expression_template = 'import %(path)s { inherit pkgs python ; }'

    def nix_expression(self):
        return self.expression_template % dict(path=self.path)


class OverridesUrl(object):
    def __init__(self, url):
        self.url = url

    expression_template = (
        'let src = pkgs.fetchurl { ' +
        'url = %(url)s; sha256 = "%(sha_string)s"; ' +
        '}; in ' +
        'import "${src}" { inherit pkgs python ; }'
    )

    def nix_expression(self):
        command = 'nix-prefetch-url {url}'.format(
            url=self.url
        )

        return_code, output = cmd(
            command, verbose=False, stderr=subprocess.DEVNULL
        )
        sha_sum = output.strip()
        if len(sha_sum) != 52 or return_code != 0:
            raise click.ClickException(
                'Could not determin hash for url %{url}s' % dict(
                    url=self.url
                )
            )
        return self.expression_template % dict(
            url=self.url,
            sha_string=sha_sum
        )


class OverridesGit(object):
    def __init__(self, repo_url, path, rev=None):
        self.repo_url = repo_url
        self.path = path
        self.rev = rev

    expression_template = (
        'let src = pkgs.fetchgit { ' +
        'url = "%(url)s"; ' +
        'sha256 = "%(sha256)s"; ' +
        'rev = "%(rev)s"; ' +
        '} ; in import "${src}/%(path)s" { inherit pkgs python; }'
    )

    def nix_expression(self):
        repo_data = prefetch_git(self.repo_url, self.rev)
        return self.expression_template % dict(
            url=repo_data['url'],
            sha256=repo_data['sha256'],
            rev=repo_data['rev'],
            path=self.path
        )


def url_to_overrides(url_string):
    url = urlparse(url_string)
    if url.scheme == '':
        return OverridesFile(url.path)
    elif url.scheme == 'file':
        return OverridesFile(url.path)
    elif url.scheme == 'http' or url.scheme == 'https':
        return OverridesUrl(url.geturl())
    elif url.scheme.startswith('git+'):
        if not url.fragment:
            raise UnsupportedUrlError(
                ('Cannot handle overrides with no path given, offeding url was'
                 ' {url}.')
                .format(
                    url=url_string
                )
            )
        fragments = dict(
            map(lambda x: x.split('='), url.fragment.split('&'))
        )
        return OverridesGit(
            repo_url=urldefrag(url.geturl()[4:])[0],
            path=fragments['path'],
            rev=fragments.get('rev', None),
        )
    else:
        raise UnsupportedUrlError('Cannot handle common overrides url %s' %
                                  url_string)


class OverridesUrlParam(click.ParamType):
    name = 'url'

    def convert(self, value, param, ctx):
        try:
            return url_to_overrides(value)
        except UnsupportedUrlError as e:
            self.fail(str(e), param, ctx)


OVERRIDES_URL = OverridesUrlParam()



import os
import sys

import click


DEFAULT_NIX = '''# generated using pypi2nix tool (version: %(version)s)
# See more at: https://github.com/garbas/pypi2nix
#
# COMMAND:
#   pypi2nix %(command_arguments)s
#

{ pkgs ? import <nixpkgs> {}
}:

let

  inherit (pkgs) makeWrapper;
  inherit (pkgs.stdenv.lib) fix' extends inNixShell;

  pythonPackages =
  import "${toString pkgs.path}/pkgs/top-level/python-packages.nix" {
    inherit pkgs;
    inherit (pkgs) stdenv;
    python = pkgs.%(python_version)s;
    # patching pip so it does not try to remove files when running nix-shell
    overrides =
      self: super: {
        bootstrapped-pip = super.bootstrapped-pip.overrideDerivation (old: {
          patchPhase = old.patchPhase + ''
            sed -i \
              -e "s|%(paths_to_remove)s|#%(paths_to_remove)s|"  \
              -e "s|%(self_uninstalled)s|#%(self_uninstalled)s|"  \
                $out/${pkgs.python35.sitePackages}/pip/req/req_install.py
          '';
        });
      };
  };

  commonBuildInputs = %(extra_build_inputs)s;
  commonDoCheck = %(enable_tests)s;

  withPackages = pkgs':
    let
      pkgs = builtins.removeAttrs pkgs' ["__unfix__"];
      interpreter = pythonPackages.buildPythonPackage {
        name = "%(python_version)s-interpreter";
        buildInputs = [ makeWrapper ] ++ (builtins.attrValues pkgs);
        buildCommand = ''
          mkdir -p $out/bin
          ln -s ${pythonPackages.python.interpreter} \
              $out/bin/${pythonPackages.python.executable}
          for dep in ${builtins.concatStringsSep " " \
              (builtins.attrValues pkgs)}; do
            if [ -d "$dep/bin" ]; then
              for prog in "$dep/bin/"*; do
                if [ -f $prog ]; then
                  ln -s $prog $out/bin/`basename $prog`
                fi
              done
            fi
          done
          for prog in "$out/bin/"*; do
            wrapProgram "$prog" --prefix PYTHONPATH : "$PYTHONPATH"
          done
          pushd $out/bin
          ln -s ${pythonPackages.python.executable} python
          ln -s ${pythonPackages.python.executable} \
              python%(python_major_version)s
          popd
        '';
        passthru.interpreter = pythonPackages.python;
      };
    in {
      __old = pythonPackages;
      inherit interpreter;
      mkDerivation = pythonPackages.buildPythonPackage;
      packages = pkgs;
      overrideDerivation = drv: f:
        pythonPackages.buildPythonPackage (drv.drvAttrs // f drv.drvAttrs // \
                                           { meta = drv.meta; });
      withPackages = pkgs'':
        withPackages (pkgs // pkgs'');
    };

  python = withPackages {};

  generated = self: {
%(generated_package_nix)s
  };
  localOverridesFile = %(overrides_file)s;
  overrides = import localOverridesFile { inherit pkgs python; };
  commonOverrides = [
%(common_overrides)s
  ];
  allOverrides =
    (if (builtins.pathExists localOverridesFile)
     then [overrides] else [] ) ++ commonOverrides;

in python.withPackages
   (fix' (pkgs.lib.fold
            extends
            generated
            allOverrides
         )
   )
'''

GENERATED_NIX = '''# generated using pypi2nix tool (version: %s)
#
# COMMAND:
#   pypi2nix %s
#

{ pkgs, python, commonBuildInputs ? [], commonDoCheck ? false }:

self: {
%s
}
'''

GENERATED_PACKAGE_NIX = '''
    "%(name)s" = python.mkDerivation {
      name = "%(name)s-%(version)s";
      src = %(fetch_expression)s;
      doCheck = commonDoCheck;
      buildInputs = commonBuildInputs;
      propagatedBuildInputs = %(propagatedBuildInputs)s;
      meta = with pkgs.stdenv.lib; {
        homepage = "%(homepage)s";
        license = %(license)s;
        description = "%(description)s";
      };
    };
'''

OVERRIDES_NIX = '''
{ pkgs, python }:

self: super: {
%s
}
'''


def main(packages_metadata,
         requirements_name,
         requirements_files,
         requirements_frozen,
         extra_build_inputs,
         enable_tests,
         python_version,
         current_dir,
         common_overrides=[],
         ):
    '''Create Nix expressions.
    '''

    default_file = os.path.join(
        current_dir, '{}.nix'.format(requirements_name)
    )
    overrides_file = os.path.join(
        current_dir, '{}_override.nix'.format(requirements_name)
    )
    frozen_file = os.path.join(
        current_dir, '{}_frozen.txt'.format(requirements_name)
    )

    version_file = os.path.join(os.path.dirname(__file__), 'VERSION')
    with open(version_file) as f:
        version = f.read()

    metadata_by_name = {x['name'].lower(): x for x in packages_metadata}

    generated_packages_metadata = []
    for item in sorted(packages_metadata, key=lambda x: x['name']):
        propagatedBuildInputs = '[ ]'
        if item.get('deps'):
            deps = [x for x in item['deps']
                    if x.lower() in metadata_by_name.keys()]
            if deps:
                propagatedBuildInputs = "[\n%s\n    ]" % (
                    '\n'.join(sorted(
                        ['      self."%s"' %
                         (metadata_by_name[x.lower()]['name'])
                         for x in deps if x not in [item['name']]]
                    )))
        fetch_type = item.get('fetch_type', None)
        if fetch_type == 'path':
            fetch_expression = './' + os.path.relpath(item['url'], current_dir)
        elif fetch_type == 'fetchgit':
            fetch_expression = 'pkgs.fetchgit { url = "%(url)s"; '\
                '%(hash_type)s = "%(hash_value)s"; rev = "%(rev)s"; }' % dict(
                    url=item['url'],
                    hash_type=item['hash_type'],
                    hash_value=item['hash_value'],
                    rev=item['rev']
                )
        elif fetch_type == 'fetchhg':
            fetch_expression = 'pkgs.fetchhg { url = "%(url)s"; '\
                '%(hash_type)s = "%(hash_value)s"; rev = "%(rev)s"; }' % dict(
                    url=item['url'],
                    hash_type=item['hash_type'],
                    hash_value=item['hash_value'],
                    rev=item['rev']
                )
        else:
            fetch_expression = 'pkgs.fetchurl { url = "%(url)s"; '\
                '%(hash_type)s = "%(hash_value)s"; }' % dict(
                    url=item['url'],
                    hash_type=item['hash_type'],
                    hash_value=item['hash_value'],
                )

        generated_packages_metadata.append(dict(
            name=item.get("name", ""),
            version=item.get("version", ""),
            fetch_expression=fetch_expression,
            propagatedBuildInputs=propagatedBuildInputs,
            homepage=item.get("homepage", ""),
            license=item.get("license", ""),
            description=item.get("description", ""),
        ))

    generated = '\n\n'.join(
        GENERATED_PACKAGE_NIX % x for x in generated_packages_metadata
    )

    overrides = OVERRIDES_NIX % ""

    common_overrides_expressions = [
        '    (' + override.nix_expression() + ')'
        for override in common_overrides
    ]

    default = DEFAULT_NIX % dict(
        version=version,
        command_arguments=' '.join(sys.argv[1:]),
        python_version=python_version,
        extra_build_inputs=(
            extra_build_inputs and
            "with pkgs; [ %s ]" % (' '.join(extra_build_inputs)) or
            "[]"
        ),
        overrides_file='.' + overrides_file[len(current_dir):],
        enable_tests=str(enable_tests).lower(),
        generated_package_nix=generated,
        common_overrides='\n'.join(common_overrides_expressions),
        paths_to_remove="paths_to_remove.remove(auto_confirm)",
        self_uninstalled="self.uninstalled = paths_to_remove",
        python_major_version=python_version.replace("python", "")[0],
    )

    if not os.path.exists(overrides_file):
        with open(overrides_file, 'w+') as f:
            f.write(overrides.strip())
            click.echo('|-> writing %s' % overrides_file)

    with open(default_file, 'w+') as f:
        f.write(default.strip())

    with open(requirements_frozen) as f:
        frozen_content = f.read()

    with open(frozen_file, 'w+') as f:
        f.write(frozen_content)



# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages

with open('src/pypi2nix/VERSION') as f:
    version = f.read().strip()

setup(
    name="pypi2nix",
    version=version,
    author=u"Rok Garbas, Cillian de Róiste, Jaka Hudoklin",
    description=(
        "A tool that generates nix expressions for your python packages, so "
        "you don't have to."
    ),
    license="BSD",
    keywords="nixos nix packaging",
    url="https://github.com/NixOS/pypi2nix",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
    ],
    entry_points = {
        'console_scripts': [
            'pypi2nix = pypi2nix.cli:main',
        ],
    },
    packages = ['pypi2nix'],
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires = [
        'click',
        'requests',
    ]
)



import click
import hashlib
import os
import shutil
import tempfile

import pypi2nix.overrides
import pypi2nix.stage0
import pypi2nix.stage1
import pypi2nix.stage2
import pypi2nix.stage3
import pypi2nix.utils


@click.command('pypi2nix')
@click.option('--version',
              is_flag=True,
              help=u'Show version of pypi2nix',
              )
@click.option('-v', '--verbose', count=True)
@click.option('-I', '--nix-path',
              envvar='NIX_PATH',
              multiple=True,
              default=None,
              help=u'Add a path to the Nix expression search path. This '
                   u'option may be given multiple times. See the NIX_PATH '
                   u'environment variable for information on the semantics '
                   u'of the Nix search path. Paths added through -I take '
                   u'precedence over NIX_PATH.',
              )
@click.option('--nix-shell',
              required=False,
              default='nix-shell',
              help=u'Path to nix-shell executable.',
              )
@click.option('--basename',
              required=False,
              default='requirements',
              help=u'Basename which is used to generate files. By default '
                   u'it uses basename of provided file.',
              )
@click.option('-C', '--cache-dir',
              required=False,
              default=None,
              type=click.Path(exists=True, file_okay=True, writable=True,
                              resolve_path=True),
              help=u'Cache directory to be used for downloading packages.',
              )
@click.option('-E', '--extra-build-inputs',
              multiple=True,
              default=None,
              help=u'Extra build dependencies needed for installation of '
                   u'required python packages.'
              )
@click.option('-N', '--extra-env',
              default='',
              help=u'Extra environment variables needed for installation of '
              u'required python packages.'
              u'Example: "LANG=en_US.UTF-8 FOO_OPTS=xyz"'
              )
@click.option('-T', '--enable-tests',
              is_flag=True,
              help=u'Enable tests in generated packages.'
              )
@click.option('-V', '--python-version',
              required=False,
              default=None,
              type=click.Choice(pypi2nix.utils.PYTHON_VERSIONS.keys()),
              help=u'Provide which python version we build for.',
              )
@click.option('-r', '--requirements',
              required=False,
              default=None,
              multiple=True,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True),
              help=u'pip requirements.txt file',
              )
@click.option('-b', '--buildout',
              required=False,
              default=None,
              type=click.Path(exists=True, resolve_path=True),
              help=u'zc.buildout configuration file',
              )
@click.option('-e', '--editable',
              multiple=True,
              required=False,
              default=None,
              type=str,
              help=u'location/url to editable locations',
              )
@click.option('-s', '--setup-requires',
              multiple=True,
              required=False,
              default=None,
              type=str,
              help=u'Extra Python dependencies needed before the installation '
                   u'to build wheels.'
              )
@click.option('-O', '--overrides',
              multiple=True,
              required=False,
              type=pypi2nix.overrides.OVERRIDES_URL,
              help=u'Extra expressions that override generated expressions ' +
                   u'for specific packages',
              )
@click.option('--default-overrides/--no-default-overrides',
              default=False,
              help=u'Apply overrides from "nixpkgs-python" (https://github.com/garbas/nixpkgs-python)', # noqa
              )
def main(version,
         verbose,
         nix_shell,
         nix_path,
         basename,
         cache_dir,
         extra_build_inputs,
         extra_env,
         enable_tests,
         python_version,
         requirements,
         buildout,
         editable,
         setup_requires,
         overrides,
         default_overrides,
         ):
    """SPECIFICATION should be requirements.txt (output of pip freeze).
    """

    if default_overrides:
        overrides += tuple([
            pypi2nix.overrides.OverridesGit(
                repo_url='https://github.com/garbas/nixpkgs-python.git',
                path='overrides.nix',
            )
        ])

    with open(os.path.join(os.path.dirname(__file__), 'VERSION')) as f:
        pypi2nix_version = f.read()

    if version:
        click.echo(pypi2nix_version)
        return

    python_versions = pypi2nix.utils.PYTHON_VERSIONS.keys()
    if not python_version:
        raise click.exceptions.UsageError(
            "Missing option \"-V\" / \"--python-version\".  Choose from " +
            (", ".join(python_versions)))

    extra_build_inputs = pypi2nix.utils.args_as_list(extra_build_inputs)
    setup_requires = pypi2nix.utils.args_as_list(setup_requires)

    # temporary pypi2nix folder and make sure it exists
    tmp_dir = os.path.join(tempfile.gettempdir(), 'pypi2nix')
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    current_dir = os.getcwd()
    requirements_name = os.path.join(current_dir, basename)

    if not cache_dir:
        cache_dir = os.path.join(tmp_dir, 'cache')

    download_cache_dir = os.path.join(cache_dir, 'download')
    wheel_cache_dir = os.path.join(cache_dir, 'wheel')
    buildout_cache_dir = os.path.join(cache_dir, 'buildout')
    pip_build_dir = os.path.join(cache_dir, 'pip')

    if not os.path.exists(download_cache_dir):
        os.makedirs(download_cache_dir)

    if not os.path.exists(wheel_cache_dir):
        os.makedirs(wheel_cache_dir)

    if not os.path.exists(buildout_cache_dir):
        os.makedirs(buildout_cache_dir)

    if not os.path.exists(pip_build_dir):
        os.makedirs(pip_build_dir)

    requirements_files = []
    if requirements:
        requirements_files += requirements

    requirements_hash = ''
    for requirements_file in requirements_files:
        requirements_hash += requirements_file
        with open(requirements_file) as f:
            requirements_hash += f.read()

    if buildout:
        requirements_hash += buildout
        with open(buildout) as f:
            requirements_hash += f.read()

    project_hash = hashlib.md5(requirements_hash.encode()).hexdigest()

    project_dir = os.path.join(tmp_dir, project_hash)
    if os.path.exists(project_dir):
        shutil.rmtree(project_dir)
    os.makedirs(project_dir)

    sources = dict()

    def handle_requirements_file(project_dir, requirements_file):

        # we find new name for our requirements_file
        new_requirements_file = "%s/%s.txt" % (
            project_dir,
            hashlib.md5(requirements_file.encode()).hexdigest(),
        )

        # we open both files: f1 to read, f2 to write
        with open(requirements_file) as f1:
            with open(new_requirements_file, "w+") as f2:
                for requirements_line in f1.readlines():
                    requirements_line = requirements_line.strip()
                    if requirements_line.startswith("-e git+") or \
                       requirements_line.startswith("-e hg+"):
                        pass
                    elif requirements_line.startswith("-e "):
                        requirements_line = requirements_line[3:]
                        try:
                            tmp_path, egg = requirements_line.split('#')
                            tmp_name = egg.split('egg=')[1]
                            tmp_path = tmp_path.strip()
                            _tmp = tmp_path.split('[')
                            if len(_tmp) > 1:
                                tmp_path = _tmp[0]
                                tmp_other = '[' + _tmp[1]
                            else:
                                tmp_other = ''
                        except:
                            raise click.ClickException(
                                "Requirement starting with `.` "
                                "should end with #egg=<name>. Line `%s` does "
                                "not end with egg=<name>" % requirements_line
                            )

                        tmp_path = os.path.abspath(os.path.join(
                            os.path.dirname(requirements_file),
                            os.path.abspath(os.path.join(
                                current_dir, tmp_path
                            )),
                        ))

                        requirements_line = "-e %s%s" % (tmp_path, tmp_other)
                        sources[tmp_name] = dict(url=tmp_path, type='path')

                    elif requirements_line.startswith("-r ."):
                        requirements_file2 = os.path.abspath(os.path.join(
                            os.path.dirname(requirements_file),
                            requirements_line[3:]
                        ))
                        new_requirements_file2 = handle_requirements_file(
                            project_dir, requirements_file2)
                        requirements_line = "-r " + new_requirements_file2
                    f2.write(requirements_line + "\n")

        return new_requirements_file

    requirements_files_tmp = []
    for requirements_file in requirements_files:
        if requirements_file in requirements:
            requirements_files_tmp.append(
                handle_requirements_file(project_dir, requirements_file))
        else:
            requirements_files_tmp.append(requirements_file)
    requirements_files = requirements_files_tmp

    click.echo('pypi2nix v{} running ...'.format(pypi2nix_version))
    click.echo('')

    if buildout:
        click.echo('Stage0: Generating requirements.txt from buildout '
                   'configuration ...')
        buildout_requirements = pypi2nix.stage0.main(
            verbose=verbose,
            buildout_file=buildout,
            project_dir=project_dir,
            buildout_cache_dir=buildout_cache_dir,
            extra_build_inputs=extra_build_inputs,
            python_version=pypi2nix.utils.PYTHON_VERSIONS[python_version],
            nix_shell=nix_shell,
            nix_path=nix_path,
            setup_requires=setup_requires,
        )
        if buildout_requirements:
            requirements_files.append(buildout_requirements)

    if editable:
        editable_file = os.path.join(project_dir, 'editable.txt')
        with open(editable_file, 'w+') as f:
            for item in editable:
                item_path = item.split('[')[0].split('#')[0]
                if item_path.startswith('.'):
                    item_path = os.path.abspath(os.path.join(current_dir, item_path))  # noqa
                if os.path.isdir(item_path):
                    f.write('-e %s\n' % item)
                else:
                    f.write('%s\n' % item)

        requirements_files = [handle_requirements_file(project_dir, editable_file)] + requirements_files  # noqa

    click.echo('Stage1: Downloading wheels and creating wheelhouse ...')

    requirements_frozen, wheels = pypi2nix.stage1.main(
        verbose=verbose,
        requirements_files=requirements_files,
        project_dir=project_dir,
        download_cache_dir=download_cache_dir,
        wheel_cache_dir=wheel_cache_dir,
        pip_build_dir=pip_build_dir,
        extra_build_inputs=extra_build_inputs,
        python_version=pypi2nix.utils.PYTHON_VERSIONS[python_version],
        nix_path=nix_path,
        setup_requires=setup_requires,
        extra_env=extra_env,
    )

    click.echo('Stage2: Extracting metadata from pypi.python.org ...')

    packages_metadata = pypi2nix.stage2.main(
        verbose=verbose,
        wheels=wheels,
        requirements_files=requirements_files,
        wheel_cache_dir=wheel_cache_dir,
        sources=sources,
    )

    click.echo('Stage3: Generating Nix expressions ...')

    pypi2nix.stage3.main(
        packages_metadata=packages_metadata,
        requirements_name=requirements_name,
        requirements_files=requirements_files,
        requirements_frozen=requirements_frozen,
        extra_build_inputs=extra_build_inputs,
        enable_tests=enable_tests,
        python_version=pypi2nix.utils.PYTHON_VERSIONS[python_version],
        current_dir=current_dir,
        common_overrides=overrides,
    )

    click.echo('')
    click.echo('Nix expressions generated successfully.')
    click.echo('')
    click.echo('To start development run:')
    click.echo('    nix-shell requirements.nix -A interpreter')
    click.echo('')
    click.echo('More information you can find at')
    click.echo('    https://github.com/garbas/pypi2nix')
    click.echo('')
